import sys
from io import BufferedReader, BytesIO
from typing import Dict, Iterator, Union, BinaryIO


# ISA has fixed width elements. These should all be equal. 103 is official sep.element position
isa_element_sep = [3, 6, 17, 20, 31, 34, 50, 53, 69, 76, 81, 83, 89, 99, 101, 103]
isa_example = "ISA*00*          *00*          *ZZ*SOMEBODYELSE   *ZZ*MAYBEYOU       *171231*2359*U*00401*000012345*0*P*:~"  # noqa


def readdocument(edi: Union[str, BinaryIO], filename='stream') -> Iterator[str]:
    """Splits text on a line_break character (unless preceeded by an escape character)."""
    if isinstance(edi, str):
        sep = detect(edi)
        edi = BytesIO(edi.encode('utf-8'))
    else:
        edi = BufferedReader(edi)
        sep = detect(str(edi.peek(110), 'ascii'))
    if not sep:
        print("Skipping...%s" % filename, file=sys.stderr)
        return

    line_break = bytes(sep['segment'], 'ascii')
    escape = bytes(sep['escape'], 'ascii') if 'escape' in sep else None
    blacklist = {b'\r', b'\n'} if sep.get('hard_wrap') else set()
    last = ''
    character = ''
    buf = []

    while True:
        character = edi.read(1)  # EOF returns b''
        if character in blacklist:
            continue
        buf.append(str(character, 'utf-8'))
        if (character == line_break and last != escape) or character == b'':
            line = "".join(buf).strip()
            buf[:] = []
            if line:
                yield line
            if character == b'':
                break
        last = character  # previously read character


def detect(text: str) -> Dict[str, str]:
    """Takes an EDI string and returns detected separators as a dictionary."""

    sep = {}
    # EDI X12: begins with ISA (106 chars) followed by a GS segment
    if text.startswith('ISA'):
        if 'GS' in text[106:110] and len(set([text[pos] for pos in isa_element_sep])) == 1:
            sep = {'element': text[103],
                   'subelement': text[104],
                   'segment': text[105],
                   'suffix': text[106:text.find('GS')]}
            if text[82] != 'U':  # X12 before repetition has 'U' here.
                sep['repetition'] = text[82]
        else:
            print("Invalid X12 ISA Header.",
                  "Expected: %s" % isa_example,
                  "Received: '%r'" % (text[:106]),
                  file=sys.stderr, sep="\n")
    # Edifact UNA: begins with UNA followed by UNB or UNG
    elif text.startswith('UNA') and 'UN' in text[3:13]:
        # ex: """UNA:+.? '\r\nUN..."""
        sep = {'subelement': text[3],
               'element': text[4],
               'release': text[6],
               'segment': text[8],
               'suffix': text[9:text.find('UN')]}
        if text[7] != ' ':  # Edifact before repetition has ' ' here.
            sep['repetition'] = text[7]
    # Edifact no UNA: begins with UNB followed by UNH
    elif text.startswith('UNB') and 'UN' in text[3:13]:
        sep = {'subelement': ':',
               'element': '+',
               'segment': "'",
               'release': '?',
               'repetition': '*',
               'suffix': text[text.find("'") + 1:text.find('UN', 3)]}
    else:
        print("Found something that doesn't look like EDI: %r" % text[:8], file=sys.stderr)

    # This detects "hard wrapped" EDI files where a CRLF is inserted every 80 chars
    # TODO: Potentially support LF (\n) only hard wrap.
    if (text[80] == text[162] == text[244] == '\r' and
        text[81] == text[163] == text[245] == '\n'):
        sep['hard_wrap'] = True
    return sep



import argparse
import os
import sys
from typing import BinaryIO, Iterable, Iterator, Tuple

import edicat
from edicat.edi import readdocument


def lprint(line: str, lineno: int, line_numbers: bool = False) -> None:
    """Print a line with optional line number"""
    if line_numbers:
        print("{0: >6}\t{1}".format(lineno + 1, line))
    else:
        print(line)


def openfiles(filenames: Iterable) -> Tuple[str, Iterator[BinaryIO]]:
    """Take an iterable of filenames and yields (filename, file object) tuples"""
    if not filenames:
        filenames = ['-']
    for filename in filenames:
        if filename == '-':
            yield 'stdin', sys.stdin.buffer
        else:
            with open(filename, 'rb') as file:
                yield filename, file


def output(filenames: Iterable, line_numbers: bool = False) -> int:
    ret_code = 0
    try:
        for filename, file in openfiles(filenames):
            lineno = 0
            for line in readdocument(file, filename):
                lprint(line, lineno, line_numbers)
            ret_code = ret_code or int(lineno == 0)
    except BrokenPipeError:
        sys.stdout = os.fdopen(1)  # suppress "Exception ignored in: [...]" when pager terminates.
    return ret_code


def main() -> None:
    parser = argparse.ArgumentParser(prog='edicat', description="Print and concatenate EDI.")
    parser.add_argument('filenames', nargs='*', help="Filename(s) or - for stdin")
    parser.add_argument("-n", "--lineno", action="store_true",
                        help="Number the output lines, starting at 1.")
    parser.add_argument('--version', action="store_true", help=argparse.SUPPRESS)
    args = parser.parse_args()
    if args.version:
        print("edicat", edicat.__version__)
    # This behavior diverges from cat, but is more user friendly.
    elif not args.filenames and sys.stdin.isatty():
        parser.print_help()
        sys.exit(1)
    else:
        sys.exit(output(args.filenames, args.lineno))


if __name__ == '__main__':
    main()



import sys
from setuptools import setup

# This check and everything above must remain compatible with Python 2.7.
if sys.version_info[:2] < (3, 5):
    raise SystemExit("Python >= 3.5 required.")

import edicat  # noqa

# Begin workaround to shave seconds off script execution.
# Issue: https://github.com/pypa/setuptools/issues/510
TEMPLATE = '''\
# -*- coding: utf-8 -*-
# EASY-INSTALL-ENTRY-SCRIPT: '{3}','{4}','{5}'
__requires__ = '{3}'
import re
import sys

from {0} import {1}

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit({2}())
'''


@classmethod
def get_args(cls, dist, header=None):
    """Monkey patch for slow stock setuptools get_args implementation."""
    if header is None:
        header = cls.get_header()
    spec = str(dist.as_requirement())
    for type_ in 'console', 'gui':
        group = type_ + '_scripts'
        for name, ep in dist.get_entry_map(group).items():
            # ensure_safe_name
            if '/' in name:
                raise ValueError("Path separators not allowed in script names")
            script_text = TEMPLATE.format(ep.module_name, ep.attrs[0],
                                          '.'.join(ep.attrs), spec, group, name)
            args = cls._get_script_args(type_, name, header, script_text)  # noqa
            for res in args:
                yield res


# Monkey patch setuptools ScriptWriter.get_args with our get_args
import setuptools.command.easy_install  # noqa
setuptools.command.easy_install.ScriptWriter.get_args = get_args
# End workaround for pypa/setuptools#510

setup(
    name='edicat',
    version=edicat.__version__,
    description='Print and concatenate EDI files',
    long_description=(
        'Detects EDI X12 and Edifact separators, printing one EDI segment '
        'per line. This allows building shell pipelines (grep, find, etc) '
        'with mixed format EDI files (X12, Edifact, CR, CR/LF, LF, no newlines '
        ' etc.) or just paging through a single EDI file.'
    ),
    url='https://github.com/notpeter/edicat',
    author='Peter Tripp',
    author_email='peter.tripp@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Text Processing',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='edi edicat edifact ansix12',
    packages=['edicat'],
    python_requires='>=3.5',
    install_requires=[],
    entry_points={
        'console_scripts': [
            'edicat = edicat.__main__:main',
        ],
    },
)
